import inspect

from werkzeug.exceptions import HTTPException
from werkzeug.routing import Map

from ponodo.helpers import import_class
from ponodo.http import Request


class ControllerDispatcher:
    def __init__(self, app):
        self.app = app

    def run(self):
        request = self.app.instances["request"]

        # todo: configurable routing registration
        from app.routes.web import routes

        routes_map = Map()
        for route in routes:
            if route.map is not None:
                route.map = None
            routes_map.add(route)

        # todo: Save routes_map to cache for faster decision
        adapter = routes_map.bind_to_environ(request.environ)

        try:
            endpoint, params = adapter.match()
        except HTTPException as e:
            return e(
                request.environ, self.app.instances["response"].werkzeug_start_response
            )

        controller, action = endpoint.split("@")
        controller = import_class(controller)
        executor = getattr(controller(app=self), action)

        arg_spec = inspect.getfullargspec(executor)
        annotations = arg_spec.annotations
        kwargs = params

        for annotation in annotations:

            if annotations[annotation] == Request:
                kwargs[annotation] = request

        # if 'request' in args:
        #     kwargs['request'] = request
        controller_result = executor(**kwargs)

        response = self.app.instances["response"]
        return response(controller_result)



import collections.abc
import json

from werkzeug.wrappers.request import Request as WerkzeugRequest
from werkzeug.wrappers.response import Response as WerkzeugResponse


class Controller:
    def __init__(self, app):
        self.app = app


class Request:
    def __init__(self, app, werkzeug_request: WerkzeugRequest):
        self.app = app
        self.werkzeug_request = werkzeug_request

    def __getattr__(self, item):
        if hasattr(self.werkzeug_request, item):
            return getattr(self.werkzeug_request, item)

    def get(self, key):
        return self.werkzeug_request.values[key]


class Response:
    def __init__(self, app, werkzeug_start_response):
        self.app = app
        self.werkzeug_start_response = werkzeug_start_response

    def __call__(self, *args, **kwargs):
        content, content_type = self._get_content_metadata(args[0])
        response = WerkzeugResponse(content, content_type=content_type, **kwargs)
        return response(
            self.app.instances["request"].environ, self.werkzeug_start_response
        )

    @staticmethod
    def _get_content_metadata(content):
        if isinstance(content, str):
            return content, "text/plain; charset=utf-8"
        elif isinstance(content, (dict, collections.abc.Sequence)):
            return json.dumps(content), "application/json; charset=utf-8"

        return content, "text/plain; charset=utf-8"



def import_class(cls: str):
    import importlib

    module_path = ".".join(cls.split(".")[:-1])
    klass = cls.split(".")[-1:][0]
    module = importlib.import_module(module_path)
    return getattr(module, klass)


def debug(*, header="", compile_flags=None, **kwargs):

    import sys

    from IPython import InteractiveShell
    from IPython.terminal.embed import InteractiveShellEmbed
    from IPython.terminal.ipapp import load_default_config

    config = kwargs.get("config")
    if config is None:
        config = load_default_config()
        config.InteractiveShellEmbed = config.TerminalInteractiveShell
        kwargs["config"] = config
    using = kwargs.get("using", "sync")
    if using:
        kwargs["config"].update(
            {
                "TerminalInteractiveShell": {
                    "loop_runner": using,
                    "colors": "NoColor",
                    "autoawait": using != "sync",
                }
            }
        )
    # save ps1/ps2 if defined
    ps1 = None
    ps2 = None
    try:
        ps1 = sys.ps1
        ps2 = sys.ps2
    except AttributeError:
        pass
    # save previous instance
    saved_shell_instance = InteractiveShell._instance
    if saved_shell_instance is not None:
        cls = type(saved_shell_instance)
        cls.clear_instance()
    frame = sys._getframe(1)
    shell = InteractiveShellEmbed.instance(
        _init_location_id="{}:{}".format(frame.f_code.co_filename, frame.f_lineno),
        colors="neutral",
        **kwargs
    )
    shell(
        header=header,
        stack_depth=2,
        compile_flags=compile_flags,
        _call_location_id="{}:{}".format(frame.f_code.co_filename, frame.f_lineno),
    )
    InteractiveShellEmbed.clear_instance()
    # restore previous instance
    if saved_shell_instance is not None:
        cls = type(saved_shell_instance)
        cls.clear_instance()
        for subclass in cls._walk_mro():
            subclass._instance = saved_shell_instance
    if ps1 is not None:
        sys.ps1 = ps1
        sys.ps2 = ps2



from abc import ABCMeta


class Container(metaclass=ABCMeta):
    def bind(self):
        raise NotImplementedError

    def singleton(self):
        raise NotImplementedError

    def instance(self):
        raise NotImplementedError

    def flush(self):
        raise NotImplementedError

    def make(self):
        raise NotImplementedError

    def resolved(self):
        raise NotImplementedError

    def resolving(self):
        raise NotImplementedError



from inflect import engine


class ModelMeta(type):

    def __getattr__(cls, item):
        return getattr(cls(), item)


class Field:

    def __init__(self, column):
        self.column = column

    def __gt__(self, other):
        return self._('>', other)

    def __ge__(self, other):
        return self._('>=', other)

    def __lt__(self, other):
        return self._('<', other)

    def __le__(self, other):
        return self._('<=', other)

    def __eq__(self, other):
        return self._('=', other)

    def __ne__(self, other):
        return self._('!=', other)

    def __contains__(self, item):
        return self.contains(item)

    def _build(self, operator, value):
        return f"({self.column} {operator} {value})"

    def like(self, value, sensitive=True):
        operator = 'LIKE' if sensitive else 'ILIKE'
        return f"{self.column} {operator} '{value}'"

    def contains(self, value, sensitive=True):
        return self.like(f"%{value}%", sensitive)

    def starts_with(self, value, sensitive=True):
        return self.like(f"{value}%", sensitive)

    def ends_with(self, value, sensitive=True):
        return self.like(f"%{value}", sensitive)

    def between(self, begin, end):
        return self._('BETWEEN', f"{begin} AND {end}")

    _ = _build


field = Field


class WhereGrammar:

    def __init__(self, wheres):
        # self._wheres = where_builder
        self.raws = wheres['raws']
        self.chunks = wheres['chunks']

    def __repr__(self):
        query = ""

        for chunk in self.chunks:
            column = chunk['column']
            operator = chunk['operator']
            value = chunk['value']
            self.raws.append(f"{column} {operator} {value}")

        query += " AND ".join(self.raws)

        return query


class QueryBuilder:

    def __init__(self, table):
        self.table = table
        self._query = ""

        self._selects = []
        # self.where_raws = []
        # self.where_chunks = []
        self._wheres = {
            'raws': [],
            'chunks': []
        }

    def __repr__(self):
        table = self.table

        columns = '*' if len(self._selects) == 0 else ", ".join(self._selects)

        self._query = f"SELECT {columns} FROM {table}"
        breakpoint()
        where_grammar = WhereGrammar(self._wheres)
        self._query += f" WHERE {where_grammar}"

        return self._query

    def all(self):
        """
        User.all()
        -> select * from users

        :return:
        """
        # return "SELECT * FROM " + self.table
        return self

    def where(self, *args, **kwargs):
        """
        User.where(name='Foo')
        -> Select * from users where name = 'Foo'

        User.where(name='Foo').where(age=10)
        -> Select * from users where name = 'Foo' and age = 10

        User.where(field(name).contains("albert"))
        User.where("albert" in field('name'))
        -> select * from users where name like '%albert%'

        User.where(field(age) > 10)
        -> select * from users where age > 10

        User.where(field('age').between(2, 39))
        User.where(2 < field('age') < 39)
        -> select * from users where age between 2 and 39

        User.where(age=[1,2,3,4,5])
        :return:
        """
        # This is mean that using raw _query
        if len(args) > 0:
            for raw in args:
                self._wheres['raws'].append(raw)

            return self

        for column in kwargs:

            operator = '='

            value = kwargs[column]
            value = self._normalize_value(value)

            if isinstance(value, list):
                operator = 'IN'
                value = kwargs[column]

            self._wheres['chunks'].append({
                'column': column,
                'operator': operator,
                'value': value
            })

        return self

    def _normalize_value(self, value):
        if isinstance(value, bool):
            value = 't' if value is True else 'f'
        if isinstance(value, str):
            value = f"'{value}'"

        if isinstance(value, (list, tuple)):
            value_temp = []
            for val in value:
                value_temp.append(str(self._normalize_value(val)))
            value = ', '.join(value_temp)
            value = f"({value})"

        return value

    def to_sql(self):
        return self._query


class Db:

    @classmethod
    def table(cls, name):
        ...


class Model(metaclass=ModelMeta):
    __table__: str = None

    def __getattr__(self, item):
        return getattr(self.builder(self.table), item)

    @property
    def builder(self):
        return QueryBuilder

    @property
    def table(self):
        if self.__table__ is None:
            table_name = self.__class__.__name__.lower()
            return engine().plural(table_name)

        return self.__table__



import click
from werkzeug import run_simple

from ponodo.core import handle_application


@click.group()
def cli():
    """Ask some command to execute nicely"""


@cli.command()
@click.option(
    "-p",
    "--port",
    default=4000,
    show_default=True,
    help="Set the development server port",
)
def serve(port):
    """Run local development server"""
    run_simple(
        "localhost",
        port,
        application=handle_application,
        use_reloader=True,
        use_debugger=True,
        reloader_type="watchdog",
    )


@cli.group()
def make():
    """The fast way to make file"""


@cli.command()
def migrate():
    """Perform database migrations"""



from werkzeug import Request as WerkzeugRequest

from ponodo.http import Request, Response
from ponodo.routing import ControllerDispatcher


class Facade:
    app = None


class Container:
    """ """

    def __init__(self):
        # Binding abstract to concrete
        self.bindings = {}

        # Instance of class or plain object
        self.instances = {}

    def make(self, abstract):
        if abstract in self.instances:
            return self.instances[abstract]


class Application(Container):
    is_routes_bound = False

    def __init__(self):
        super(Application, self).__init__()

    def set_werkzeug_request(self, request):
        self.instances["request"] = Request(app=self, werkzeug_request=request)
        return self

    def set_werkzeug_response(self, response):
        self.instances["response"] = Response(
            app=self, werkzeug_start_response=response
        )
        return self

    def handle(self):
        return ControllerDispatcher(app=self).run()


def handle_application(environ, start_response):
    application = Application()
    application.set_werkzeug_request(WerkzeugRequest(environ))
    application.set_werkzeug_response(start_response)
    return application.handle()



import re

from werkzeug.routing import Rule


class RuleGenerator:
    def _new_rule(self, path, **kwargs):
        rule_kwargs = {
            "string": self._make_path(path),
            "endpoint": self._make_endpoint(**kwargs),
            "methods": self._make_methods(**kwargs),
        }

        return Rule(**rule_kwargs)

    @staticmethod
    def _make_path(path: str):
        """
        Transform
        - `@argument:int` into `<int:argument>`
        - `@argument:string` into `<string(length=2):argument>`
        """
        if not path.endswith("/"):
            path = f"{path}/"
        return re.sub(r"@([a-z]+)((:?)([a-z0-9()=]+)?)", r"<\4\3\1>", path)

    @staticmethod
    def _make_methods(**kwargs):
        methods = kwargs["methods"]
        if isinstance(methods, str):
            return [methods]
        return methods

    @staticmethod
    def _make_endpoint(**kwargs):
        controllers_location = "app.http.controllers"

        if "to" in kwargs:
            # Transform `foo@bar` into full path of the controller location,
            # and it's action. For example the sample above will transform into
            # `app.http.controllers.foo_controller.FooController@bar`

            to = kwargs["to"]

            # if "." in to:
            #     # It will transform `foo.bar@baz` into
            #     # `app.http.controllers.foo.bar_controller.BarController@baz`
            #
            #     modules = to.split(".")
            #
            #     return f"{controllers_location}.{modules[0]}"
            # debug()
            # from IPython import embed
            # embed()

            controller_candidate, action = to.split("@")
            controller_module_name = f"{controller_candidate}_controller"
            class_name = controller_module_name.title().replace("_", "")

            return (
                f"{controllers_location}.{controller_module_name}.{class_name}@{action}"
            )

        if "controller" in kwargs:
            module = kwargs["controller"].__module__
            class_name = kwargs["controller"].__name__
            endpoint = f"{module}.{class_name}"

            if "action" in kwargs:
                action = kwargs["action"]
                return f"{endpoint}@{action}"

            return endpoint

    def get(self, path, **kwargs):
        return self._new_rule(path, methods="GET", **kwargs)

    def post(self, path, **kwargs):
        return self._new_rule(path, methods="POST", **kwargs)

    def put(self, path, **kwargs):
        return self._new_rule(path, methods="PUT", **kwargs)

    def patch(self, path, **kwargs):
        return self._new_rule(path, methods="PATCH", **kwargs)

    def delete(self, path, **kwargs):
        return self._new_rule(path, methods="DELETE", **kwargs)


class RouteMeta(type):

    def __getattr__(cls, item):
        return getattr(cls(), item)


class Route(metaclass=RouteMeta):

    def __getattr__(self, item):
        return getattr(self.generator(), item)

    @property
    def generator(self):
        return RuleGenerator


get = Route.get
post = Route.post
put = Route.put
patch = Route.patch
delete = Route.delete



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ponodo', 'ponodo.abc', 'ponodo.console', 'ponodo.database', 'ponodo.routing']

package_data = \
{'': ['*']}

install_requires = \
['Werkzeug>=2.2.1,<3.0.0',
 'click>=8.1.3,<9.0.0',
 'inflect>=6.0.0,<7.0.0',
 'ipython>=8.4.0,<9.0.0']

setup_kwargs = {
    'name': 'ponodo',
    'version': '0.1.0a1',
    'description': 'Web Application Framework for Pythonista',
    'long_description': None,
    'author': 'Fathur Rohman',
    'author_email': 'hi.fathur.rohman@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
